NAME="$0"
BASE="${NAME%.sh}"
ONLY="../dist/build/only/only"
GREP="grep"
HEAD="head"
TAIL="tail"
CUT="cut"

cd "$(dirname $NAME)"

function test-eq () {
    if [ x"$1" = x"$2" ] ; then
	echo -n PASS
	echo " : $BASE"
    else
	echo -n FAIL
	echo " : $BASE"
	echo "---"
	echo "$1"
	echo "+++"
	echo "does not equal:"
	echo "---"
	echo "$2"
	echo "+++"
	echo
    fi
}
